import socket
import netifaces
import struct
from threading import Lock,Thread
import paho.mqtt.client as mqtt
import mmdvminput
import mmdvmoutput
import prepareSMS
import voiceclips
import nmea
import json

UDP_PORT   = 62031   # Port to MMDVM
senderID   = 2420001 # Sender DMR ID
#receiverID = 24203   # Receiver DMR ID
routeID    = 24299   # TG for IP routing through DMR
MQTTBROKER = '127.0.0.1'

# We interface to MMDVMs, so we need to UDP sockets
udpsock1 = socket.socket(socket.AF_INET,socket.SOCK_DGRAM) # This is for receiving data
udpsock2 = socket.socket(socket.AF_INET,socket.SOCK_DGRAM) # This is for sending data to DMR
rawsockin = socket.socket(socket.AF_PACKET,socket.SOCK_RAW,socket.htons(0x800))
rawsockout = socket.socket(socket.AF_INET,socket.SOCK_RAW,socket.IPPROTO_RAW)
udpsock1.bind(('10.10.1.2', UDP_PORT))
udpsock2.bind(('10.10.2.2', UDP_PORT))

macaddrs = list()  # List of own MAC addreses
ipv4addrs = list() # List of own IP adresses
mqttclient = mqtt.Client("DMRgateway")
mutex = Lock() # Used for thread synchronization

def ip2int(addr):
    return struct.unpack("!I", socket.inet_aton(addr))[0]

def int2ip(addr):
    return socket.inet_ntoa(struct.pack("!I", addr))


# Return a list of IP addresses on the interfaces in integer format
def ipv4_addresses():
	ipaddr = list()
	for i in netifaces.interfaces():
		addrs = netifaces.ifaddresses(i)
		if netifaces.AF_INET in addrs:
			for a in addrs[netifaces.AF_INET]:
				ipaddr.append(ip2int(a['addr']))
	return ipaddr

# Return a list of MAC addresses on the interfaces in integer format
def mac_addresses():
	macaddr = list()
	for i in netifaces.interfaces():
		addrs = netifaces.ifaddresses(i)[netifaces.AF_LINK]
		mac = addrs[0]['addr']
		b = bytes.fromhex(mac.replace(':',''))
		macaddr.append(int(mac.replace(':',''),16))
	return macaddr

# Now call these methods only once
ipv4addrs = ipv4_addresses()
macaddrs = mac_addresses()

# Install a handler for received SMS
sms_topic = "messages/fromradios/tg/"
def processSMS(src,tg,text):
	print("GWTEST: SMS received from %d (tg %d): %s" % (src,tg,text))
	# publish to mqtt
	public = {"TG":tg,"SRC":str(src),"TEXT":text}
	mqttclient.publish(sms_topic+str(tg),json.dumps(public))

nmea_topic = "location/radio/"
def processNMEA(src,nmeastr):
	global nmea_topic
	print ("GWTEST: NMEA report from %d : %s " % (src,nmeastr))
	# publish to mqtt
	locreport = nmea.longNMEA(nmeastr)
	print ("NMEA locreport",locreport)
	mqttclient.publish(nmea_topic+str(src),json.dumps(locreport))




MMDVM_ADDR = ('127.0.0.1',1000) # Just to avoid runtime errors

mmdvminput.setSMShandler(processSMS)
mmdvminput.setNMEAhandler(processNMEA)

def udp_listen(udpsock,nr): # Data input from MMDVM
	global MMDVM_ADDR
	while True:
		(message, addr) = udpsock.recvfrom(2024)
		print("Received UDP nr %d message:"%nr,message)
		MMDVM_ADDR = addr
		response = mmdvminput.mmdvminput(message)
		# Write response back to socket (need to syncronize)
		if response != None :
			mutex.acquire()
			for r in response:
				udpsock.sendto(r,addr)
			mutex.release()

def raw_listen():
	while True:
		ethframe = rawsockin.recv(2000)
		dstmac = int.from_bytes(ethframe[:6],"big")
		srcmac = int.from_bytes(ethframe[6:12],"big")
#		if srcmac in macaddrs:	continue # Frames coming from myself
		if dstmac & 0x010000000000: continue # Multicast frames

		ippacket = ethframe[14:]
		proto	 = ippacket[10:12]
		srcip	 = ippacket[12:16]
		dstip    = ippacket[16:20]
		dstip_num = int.from_bytes(dstip,"big") # Convert to number
#		print ("ETHFRAME RECEIVED, ipdst=",int2ip(dstip_num),hex(dstmac))
		if dstip_num in ipv4addrs: continue # Packets addressed to self
		if (dstip_num > (0b1110 << 28)): continue # Multicast IP
		# The packet is approved for passing to the DMR link
		print ("IP packet for routing: ",ippacket.hex())
		mmdvmpdu = mmdvmoutput.makeMMDVMpdu(ippacket,senderID,routeID)
		mutex.acquire()
		for x in mmdvmpdu:
			udpsock2.sendto(x,MMDVM_ADDR) # Send through unit B
		mutex.release()

def on_mqtt_message(client,userdata,recvdata):
		global MMDVM_ADDR
		# Receive publication from mosquitto, prepare SMS and send
		topic = recvdata.topic
		jsonmsg = recvdata.payload.decode("utf-8")
		message = json.loads(jsonmsg)
		print("From MQTT:", message,topic)
		if "TEXT" in message:
			text = message['TEXT']
			receiverID = int(message['DST'])
			ippacket = prepareSMS.prepareSMS(text,senderID,receiverID)
			mmdvmpdu = mmdvmoutput.makeMMDVMpdu(ippacket,senderID,receiverID)
		elif "VOICE" in message:
			voice = int(message['VOICE'])
			receiverID = int(message['DST'])
			mmdvmpdu = voiceclips.getVoiceData(voice,senderID,receiverID)
		mutex.acquire()
		for x in mmdvmpdu:
			print(x.hex())
			udpsock.sendto(x,MMDVM_ADDR)
		mutex.release()

# IP packets which are not SMS should be routed out through the raw socket
def routeIP(ippacket):
	dstIP = int2ip(int.from_bytes(ippacket[16:20],"big"))
	print ("IP-pakke videresendes til ",dstIP)
	rawsockout.sendto(ippacket,(dstIP,0))

mmdvminput.setIProuter(routeIP)

x1 = Thread(target=udp_listen,args=(udpsock1,1))
x1.start()

x2 = Thread(target=udp_listen,args=(udpsock2,2))
x2.start()

x3 = Thread(target=raw_listen)
x3.start()

# Start mqtt thread here
mqttclient.connect(MQTTBROKER)
mqttclient.on_message = on_mqtt_message
mqttclient.subscribe("messages/toradios/unit/+") # Until further
#mqttclient.subscribe("#") # Until further
mqttclient.loop_start()

print("Initialization complete")
