from hashlib import sha256
from dmr_utils3.decode import *
from bitarray import bitarray
from crc import crc32DMR,crc_poly
import voiceclips as vc
#from dmr_utils3 import decode

##########################################
# DMR Modem function
##########################################

repeaterID = b'\00\x24\xed\xc2'
server_id = b'\x6c\xbf\x80\xa5'
#server_id = b'\x6c\xbf\x80\xa5'
approved_repeaters = [b'\00\x24\xed\xdf',b'\00\x24\xed\xc2'] # List of allowed id
credentials = b'passord'

# This is a counter to assemble a series of data blocks. When 0, the buffer is
# sent to analysis and the buffer is cleared. The value is set by a block
# with datatype 6 and decremented by blocks with datatype 7.
blockstofollow = 0
dt7buf = bytes()
poc = 0

def defaultIPhandler(ippacket,srcID,dstID):
	# Is this a UDP packet?
	srcNr = int.from_bytes(ippacket[13:16],"big")
	dstNr = int.from_bytes(ippacket[17:20],"big")
	dstLen = len(str(dstNr))
	if ippacket[9] == 17 and ippacket[20:24] == b'\x13\x98\x13\x98': #DMR standard
		if dstLen < 7:
			print("Probably DMR-SMS")
			processSMS(srcNr,dstNr,ippacket[32:].decode('utf-16le'))
		else:
			print("SMS, but to an individual, ignored")
	elif ippacket[9] == 17 and ippacket[20:24] == b'\x0f\xa7\x0f\xa7': #M-SMS
		if dstLen < 7:
			print("Probably M-SMS")
			processSMS(srcNr,dstNr,ippacket[38:].decode('utf-16le'))
		else:
			print("SMS, but to an individual, ignored")
	elif ippacket[5:9] ==  b'\x27\x7e\x27\x7e':
		srcNr = int.from_bytes(srcID,"big")
		nmea = str(ippacket[9:],"utf-8")
		processNMEA(srcNr,nmea)
	else:
		print ("Probably normal IP, pass on to router")
		# TODO check if TG is the assigned routerID
		processIProuting(ippacket)

def defaultSMShandler(src,tg,text):
	print("SMS received from %d (tg %d): %s" % (src,tg,text))

def defaultIProuting(ippacket):
	print("IP received: ",ippacket)

def defaultNMEAhandler(src,nmea):
	print("NMEA report from %d : %s " % (src,nmea))

processIPpacket = defaultIPhandler
processSMS = defaultSMShandler
processIProuting = defaultIProuting
processNMEA = defaultNMEAhandler

def setIPhandler(handler):
	global processIPpacket
	processIPpacket = handler

def setSMShandler(handler):
	global processSMS
	processSMS = handler

def setIProuter(handler):
	global processIProuting
	processIProuting = handler

def setNMEAhandler(handler):
	global processNMEA
	processNMEA = handler

# Data from MMDVM is a mix of control and user traffic.
# This function handles the control dialog and passes
# user traffic on to processing. The result is an IP
# packet extracted from incoming bursts.
def mmdvminput(data):
	global repeaterID
	# Authentication, heartbeat and config info protocol.
	if (data[:4] == b'RPTL'):
		# Sjekk også om id er godkjent
		repeaterID = data[4:]
		print(repeaterID)
		if True: # Override next line
#		if (repeaterID in approved_repeaters):
			return [b'RPTACK' + server_id]
		else:
			return [b'RPTNAK' + data[4:]]

	if (data[:4] == b'RPTK'):
		if True: #(data[4:8] in approved_repeaters):
			pass1 = data[8:40]
			pass2 = sha256(server_id+credentials).digest()
			if (pass1==pass2):
				return [b'RPTACK' + repeaterID]
			else:
				return [b'RPTNAK' + data[4:8]]
		else:
			return [b'RPTNAK' + data[4:]]

	if (data[:4] == b'DMRC'):
		# Message from MMDVM to DMRgateway
		return [b'DMRP']

	if (data[:4] == b'RPTC'):
		# Here comes the config block from Pi-star. Unpack and use of necessary
		print("Config data from MMDVM received")
		return [b'RPTACK' + repeaterID]

	# pin-pong liveness control
	if (data[:7] == b'RPTPING'):
		return [b'MSTPONG' + data[7:]]

	# DMR data received, deode and process
	if (data[:4] == b'DMRD'):
		decodeDMRdata(data)
		return None

recording = False # True while we are recording received voice
# Functions for processing packet coming from the DMR Modem
def decodeDMRdata(data):
	# Outer layer of the pdu is the MMDVM procotol elements
	global dt7buf, blockstofollow, poc, recording
	sequence = data[4]
	sourceID = data[5:8]
	destinationID = data[8:11]
	destinationNR = int.from_bytes(destinationID,"big")
	repeaterID = data[11:15]
	sigbyte = data[15]
	if (sigbyte & 0x80):slot = 2
	else: slot = 1
	singleAddress = (sigbyte & 0x40 == 0x40) # Single or group receiver
	dataFrame = (sigbyte & 0x30 == 0x20) # What we are mostly interested in
	#voiceFrameorSync  = (sigbyte & 0x10 == 0) # Yes voice, but we do not process that
	#voiceSync   = (sigbyte & 0x30 == 0x10)
	dataType = sigbyte & 0x0f # We are interested in 3 (CSBK preamble), 6 (Data header) and 7 (data block)
	streamId = data[16:20] # We do not know what to use this for. Seems insignificant
	DMRdata = data[20:53]  # This is the Trellis coded burst that we need to crunch

	# We are interested in data frames, skip voice frames for now
	if (dataFrame):
		if (dataType==1 and destinationNR in range(8000,9000)): 
			print("Recording startet")
			recording = True
			vc.createFile(destinationNR)
			vc.appendVoiceData(data)
		if (dataType==2 and recording):
			print("Recording stopped")
			recording = False
			vc.appendVoiceData(data)
			vc.closeFile()
			return None #vc.getVoiceData(3,24202,24201)
		if (dataType==3):
			# Sometimes we need to look at the CSBK preamble, but they contain
			# no info which is not in the data header (dt=6), Uncomment if necessary
			dburst = data_burst(DMRdata)
		if (dataType==6): # Data header
			dburst = data_burst(DMRdata) # Unpack and return 12 bytes of plaintext data
			header = parseType6(dburst['LC']) # Link control, 12 bytes of useful info
			blockstofollow = header['BLOCKSTOFOLLOW'] # We need to know how many data blocks follow
			poc = header['POC'] # and how many padded octets are at the end
			dt7buf = bytes() # Prepare to assemble a series of data blocks
		if (dataType==7): # contains fragments of an IP packet.
			dburst = data_burst(DMRdata) # Unpack and return 12 bytes of plaintext data
			dt7buf += dburst['LC'] # 12 bytes of plain IP data, decode when all are received
			blockstofollow -= 1
			if (blockstofollow == 0): # All blocks gathered, now analyze
				# Check CRC32 first
				crc0 = crc32DMR(dt7buf[:-4]) 	# Calculated
				crc1 = dt7buf[-4:]		# Received
				if (crc0 != crc1): 
					print("Checksum NOT OK") # Discard the buffer
					return None # Drop further indata processing
				# Now remove padding octets and parse the IP packet
				ippacket = dt7buf[:-(poc+4)]
#				print ("IP received:",ippacket.hex())
				processIPpacket(ippacket,sourceID,destinationID)
	else:
		if (recording):
			vc.appendVoiceData(data)


# Parse the data header block, which contains necessary
# information. We assume that this is an unconfirmed
# header block. Its structure is shown in clause 8.2.1.1 
# (page 76) in ETSI TS 102 361-1-V.2.5.1
def parseType6(headerblock):
	global blockstofollow, poc
	# We assume for now that this an unconfirmed header block
	byte0 = headerblock[0]
	byte1 = headerblock[1]
	dpf = byte0 & 0xf # Expected to be 2 for unconfirmed data
	sap = byte1 >> 4 # Expected to be 4 (IP)
	poc = (byte0 & 0x10) | (byte1 & 0xf)
	dst = headerblock[2:5] # 3 bytes DMR id
	src = headerblock[5:8]
	blockstofollow = headerblock[8] & 0x7f # Number of data blocks following
	# Now check crc
	crc1 = crc_poly(headerblock[0:10],16,0x1021,xor_out=0x3333) # 0x3333 is 0xCCCC ^ 0xFFFF
	crc2 = int.from_bytes(headerblock[10:12],"big")
	# We really should do someting if the crc is not ok, but how many
	# Data blocks should we discard if we don't trust the
	# "blockstofollow" value ?
	return {"DPF":dpf, "SAP":sap, "POC":poc, "DST":dst, "SRC":src, 
		"BLOCKSTOFOLLOW": blockstofollow,"CRC":(crc1==crc2)}


# Isolate the slot type and sync field in the burst, and decode the payload
# The burst format is described in fig.6.5 in ETSI ... document.
# The payload is 196 bits, split in two by a 20 bit SlotType field with
# a 48 bits sync field in the middle. The payload (called "LC" in the code)
# Is FEC-protected with Trellis coding. The payload must be unpacked, and the
# slot type fields extracted
def data_burst(_string):
#    	print(_string.hex())
	burst = to_bits(_string)
	info = burst[0:98] + burst[166:264]
	slot_type = burst[98:108] + burst[156:166]
	sync = burst[108:156] # Indicates the nature of the payload, see table 9.2 in ETSI ... document
	lc = xdecode_full_lc(info).tobytes() # Link Control, in our context this is IP fragments
	cc = to_bytes(slot_type[0:4])
	dtype = to_bytes(slot_type[4:8])
	ret = {'LC': lc.hex(), 'CC': cc, 'DTYPE': dtype, 'SYNC': sync, 'STYPE':slot_type}
	print(ret)
	return {'LC': lc, 'CC': cc, 'DTYPE': dtype, 'SYNC': sync, 'STYPE':slot_type}

# This is to decode the payload from the BPTC(96,196) FEC coding. Code taken from dmt_utils3 package
def xdecode_full_lc(_data):
    binlc = bitarray(endian='big')
    binlc.extend([_data[136],_data[121],_data[106],_data[91], _data[76], _data[61], _data[46], _data[31]])
    binlc.extend([_data[152],_data[137],_data[122],_data[107],_data[92], _data[77], _data[62], _data[47], _data[32], _data[17], _data[2]  ])
    binlc.extend([_data[123],_data[108],_data[93], _data[78], _data[63], _data[48], _data[33], _data[18], _data[3],  _data[184],_data[169]])
    binlc.extend([_data[94], _data[79], _data[64], _data[49], _data[34], _data[19], _data[4],  _data[185],_data[170],_data[155],_data[140]])
    binlc.extend([_data[65], _data[50], _data[35], _data[20], _data[5],  _data[186],_data[171],_data[156],_data[141],_data[126],_data[111]])
    binlc.extend([_data[36], _data[21], _data[6],  _data[187],_data[172],_data[157],_data[142],_data[127],_data[112],_data[97], _data[82] ])
    binlc.extend([_data[7],  _data[188],_data[173],_data[158],_data[143],_data[128],_data[113],_data[98], _data[83]])
    binlc.extend([ _data[68],_data[53],_data[174],_data[159],_data[144],_data[129],_data[114],_data[99],_data[84],_data[69],_data[54],_data[39],
    _data[24],_data[145],_data[130],_data[115],_data[100],_data[85],_data[70],_data[55],_data[40],_data[25],_data[10],_data[191]])
    return binlc


#lx = vc.getVoiceData(3,2420191,2420191)
#le = bytes(lx[0])
#lc = le[20:53]
#print(data_burst(lc))
