# Module for sending an IP packet over a DMR link
from dmr_utils3.decode import *
from bitarray import bitarray
from crc import crc32DMR,crc_poly
from dmr_utils3 import decode,golay
from random import randint

server_id = b'\x6c\xbf\x80\xa5'
#server_id = b'\x6c\xbf\x80\xa5'
approved_repeaters = [b'\00\x24\xed\xdf',b'\00\x24\xed\xc2'] # List of allowed id
credentials = b'passord'
noOfCSBKblocks = 0
slot = 2
print("Transmitting in slot ",slot," (MUST BE SLOT 2)")

# Sync pattern for BS sourced data
#framesync = to_bits(b'\xdf\xf5\x7d\x75\xdf\x5d')
#sync pattern for MS sourced data
#framesync = to_bits(b'\xd5\xd7\xf7\x7f\xd7\x57')
# Sync pattern for direct tdma data slot 1
#framesync = to_bits(b'\xf7\xfd\xd5\xdd\xfd\x55')
# Sync pattern for direct tdma data slot 2
framesync = to_bits(b'\xd7\x55\x7f\x5f\xf7\xf5')

def makeMMDVMpdu(ippacket,srcID,dstID):
	streamID = randint(10000,99999) # Just make up a number
	group = len(str(dstID)) < 7 # Indiviual IDs have 7 digits
	datablocks,noOfPocs = makeDataBlocks(ippacket)
	headerblock = makeHeaderBlock(len(datablocks),noOfPocs,srcID,dstID,group)
	CSBKpreambles = makeCSBKpreamble(len(datablocks),srcID,dstID,group,noOfCSBKblocks)
	MMDVMpdu = makeTransmission(CSBKpreambles,headerblock,datablocks, \
		srcID,dstID,slot,streamID,group)
	return MMDVMpdu

def makeDataBlocks(ippacket):
	#print(ippacket.hex())
	# How many data blocks are generated? 12 bytes per block + 4 byte CRC
	noOfBlocks = int((len(ippacket)+15)/12) # add 1 to truncate to roof
	# Calculate number of paddingOctets
	noOfPocs = noOfBlocks*12 - 4 - len(ippacket)
	# Calculate the crc of the ippacket, excluding the padding octets CHECK THAT THIS IS CORRECT
	crcValue = crc32DMR(ippacket+noOfPocs*(b'\0'))
	# Now split the IP packet in a list of data blocks
	datablocks = list()
	for a in range(noOfBlocks-1):
		last = min((a+1)*12,len(ippacket))
		pad = (a+1)*12 - last
		datablocks.append(ippacket[a*12:last]+pad*b'\0')
	# Fill in last
	if (noOfPocs < 8): datablocks.append(ippacket[last:]+noOfPocs*(b'\0')+crcValue)
	else: datablocks.append(8*b'\0'+crcValue)
#	print("NoOfBlocks:",noOfBlocks)
#	print("NoOfPocs:",noOfPocs)
#
#	for a in datablocks:
#		print(a.hex())
	return (datablocks,noOfPocs)

# Make unconfirmed packet header for IP transport
def makeHeaderBlock(noOfBlocks,noOfPocs,srcID,dstID,group):
	if group:
		headerBlock = b'\x82' # Group address, DPF=2 (unconfirmed)
	else:
		headerBlock = b'\x02'
	headerBlock += (noOfPocs+0x40).to_bytes(1,"big")
	headerBlock += dstID.to_bytes(3,"big") # 3 bytes
	headerBlock += srcID.to_bytes(3,"big")
	headerBlock += (0x80+noOfBlocks).to_bytes(1,"big")
	headerBlock += b'\0'
	# Make crc
	crcValue = crc_poly(headerBlock,16,0x1021,xor_out=0x3333)
	headerBlock += crcValue.to_bytes(2,"big")
	print("HeaderBlock:",headerBlock.hex())
#	print(hex(crc_poly(headerBlock,16,0x1021,xor_out=0x56f6)))
	return headerBlock

def makeCSBKpreamble(noOfBlocks,srcID,dstID,group,howMany=1):
	preambles = list()
	if group:
		preamble = b'\xbd\x00\xc0'
	else:
		preamble = b'\xbd\x00\x80'
	srcB = srcID.to_bytes(3,"big")
	dstB = dstID.to_bytes(3,"big")
	while howMany>0:
		blockstofollow = (howMany+noOfBlocks).to_bytes(1,"big")
		p0 = preamble + blockstofollow + dstB + srcB
		crcValue = crc_poly(p0,16,0x1021,xor_out=0x5a5a)
		p0 += crcValue.to_bytes(2,"big")
		preambles.append(p0)
		howMany -= 1
	return preambles

# Turn the lists of blocks into bursts and wrp them in MMDVM PDUs

def makeTransmission(csbk,header,data,srcID,dstID,slot,streamID,group):
	transmission = list()
	seqNr = 0;
	for x in csbk:
		transmission.append(makeMMDVM(x,3,slot,streamID,seqNr,srcID,dstID,group))
		seqNr += 1
	transmission.append(makeMMDVM(header,6,slot,streamID,seqNr,srcID,dstID,group))
	seqNr += 1
	for x in data:
		transmission.append(makeMMDVM(x,7,slot,streamID,seqNr,srcID,dstID,group))
		seqNr += 1
	return transmission

# Take one block, turn them into a burst and make one MMDVM PDU
def makeMMDVM(block,datatype,slot,streamID,seqNr,srcID,dstID,group):
	repeaterID = 2420191 # TODO make configurable
	burst = makeBurst(block,datatype)
	slfc = 0x20 | datatype
	if slot==2: slfc = slfc | 0x80
	if not group: slfc = slfc | 0x40
	mmdvm = b'DMRD' + seqNr.to_bytes(1,"big") + srcID.to_bytes(3,"big") + dstID.to_bytes(3,"big") + \
		repeaterID.to_bytes(4,"big") + slfc.to_bytes(1,"big") + streamID.to_bytes(4,"big") + burst + b'\0\0'
	return mmdvm

# Take one block, turn them into a burst (Trellis coded and so on)
def makeBurst(block,datatype,cc=1):
	if (datatype not in [3,6,7]): return None
	#print("Blocks:",block.hex())
	b0 = bptc.encode_19696(block)
	b1 = bptc.interleave_19696(b0)
	slottype = (cc << 4 | datatype).to_bytes(1,'big')
	fecparity = to_bits(golay.encode_2087(slottype).to_bytes(3,'big'))[4:]
	block = b1[:98] + fecparity[:10] + framesync + fecparity[10:] + b1[98:]	
	#print("Encoded:",block.tobytes().hex())
	return block.tobytes()
