from bitarray import bitarray
import time

#nmeashort = bytes.fromhex("700f48141429a50cc0d2b1e2")

def toint(ba):
	x = 0
	y = ba.tolist()
	for i in y:
		x = x*2 + i
	return x

# Parse the bits from NMEA short form (IEC-61162-1)
def shortNMEA(nmeabytes):
	nb = bitarray()
	nb.frombytes(nmeabytes)
	c     = nb[0]		# 1 - encrypted
	ns    = nb[1]		# 1 - north lat
	ew    = nb[2]		# 1 - east lon
	q     = nb[3]		# 1 - GPS fix
	speed = toint(nb[4:11])	# in knots
	ndeg  = toint(nb[11:18]) # lat degrees
	nmin  = toint(nb[18:24]) # lat min (0-59)
	nminf = toint(nb[24:38]) # lat min fraction
	edeg  = toint(nb[38:46]) # lon degrees
	emin  = toint(nb[46:52]) # lon minutes
	eminf = toint(nb[52:66]) # lon min fraction
	utchh = toint(nb[66:71]) # utc hours
	utcmm = toint(nb[71:77]) # utc minutes
	utcss = toint(nb[77:83]) # utc seconds
	lon = edeg + (emin + (eminf/1000.0))/60.0
	lat = ndeg + (nmin + (nminf/1000.0))/60.0

	if ns==0: lat = -lat
	if ew==0: lon = -lon

	# Some GPS receivers (TYT 2017ND) has been observed to return
	# utchh=0 always. For this reason, we need to supply the computer
	# clock data instead
	if (utchh == 0): 
		inthh = time.gmtime()[3]
		intmm = time.gmtime()[4]
		# If there is a large difference on minute data, we might
		# be on the edge between two hour values
		if (utcmm > intmm + 50): utchh = inthh - 1 # Alloww 10 min drift
		elif (intmm > utcmm + 50): utchh = inthh + 1
		else: utchh = inthh

	utc = "%02d:%02d:%02d"%(utchh,utcmm,utcss)
	return {"LAT":lat,"LON":lon,"UTC":utc}

#print (shortNMEA(nmeashort))


nmealong = "$GPRMC,120416.000,A,6108.12223,N,01026.52411,E,0.00,5.03,240322,,,A*7A"
# Parse the long from for NMEA data, must start with $GPRMC
def longNMEA(nmeastring):
	elements = nmeastring.split(",")
	if elements[0] != "$GPRMC": return
	if elements[2] != 'A': return # Warning from receiver

	latdeg = int(float(elements[3]) / 100)
	latmin = float(elements[3]) % 100
	lat    = latdeg + (latmin/60.0)
	if elements[4] == 'S': lat = -lat

	londeg = int(float(elements[5]) / 100)
	lonmin = float(elements[5]) % 100
	lon    = londeg + (lonmin/60.0)
	if elements[6] == "W": lon = -lon

	tm = elements[1]
	utc = tm[0:2] + ":" + tm[2:4] + ":" + tm[4:6]

	return {"LAT":lat,"LON":lon,"UTC":utc}

#print(longNMEA(nmealong))
