# Take a SMS payload and make a IP/UDP  segment.
# IP addresses are sent as parameters for the sake of checksum generation,
# we already know the length and protocol value (17, UDP)


# SMS payload contains some prefix bytes which we do not what means
# so we simply copy them and hops for the best. We may want to have a
# collection of these prefixes from different radios in order to
# solve interoperabiity problems
prefixAnyTun = b'\x00\r\x00\n' 
SMS_PORT   = 5016    # UDP port used for SMS



def prepareSMS(message,srcID,dstID,prefix=prefixAnyTun):
	# Make IP addresses from srcID and dstID
	srcIP = b'\x0c'+ srcID.to_bytes(3,"big")
	group = (len(str(dstID)) < 7) # Individual IDs has 7 digit
	if group:
		dstIP = b'\xe1'+ dstID.to_bytes(3,"big") # Group addresses need class D IP address
	else:
		dstIP = b'\x0c'+ dstID.to_bytes(3,"big") # Class A address

	udpsegment = makeUDP(prefix + message.encode("UTF-16LE"),srcIP,dstIP,SMS_PORT,SMS_PORT)
	ippacket = makeIP(udpsegment,srcIP,dstIP)
	return bytes(ippacket)

# Simple checksum algorithm for IP and UDP packets
# Parameter is a byte array. Craches if buffer is odd length
def checksum(data):
        s = 0
        it = iter(data)
        for i in it:
                i = i*256+next(it)
                s += i
                s = (s & 0xffff) + (s >> 16)
        return (~s & 0xffff)

# Take a payload and port numbers, make a UDP segment
def makeUDP(payload,srcIP,dstIP,srcPort,dstPort):
	plen = (len(payload)+8).to_bytes(2,"big") # Payload length
	sp = srcPort.to_bytes(2,"big")
	dp = dstPort.to_bytes(2,"big")
	if (len(payload) % 2 == 1): payload += b'\x00' # Make segment lengt even number
	udpsegment = bytearray(sp + dp + plen + b'\x00\x00' + payload)
	# Calculate checksum and fill in header bytes
	checksumValue = checksum(udpsegment+srcIP+dstIP+plen+b'\x00\x11')
	udpsegment[6:8] = checksumValue.to_bytes(2,"big")
	return udpsegment

def makeIP(segment,srcIP,dstIP):
	plen = (len(segment)+20).to_bytes(2,"big")
	# protocol 17 is assumed, also assume other values to be current (TTL=64, etc.)
	iphdr = bytearray(b'\x45\x00'+plen+b'\x00\x03\x00\x00\x40\x11\x00\x00'+srcIP+dstIP)
	checksumValue = checksum(iphdr)
	iphdr[10:12] = checksumValue.to_bytes(2,"big")
	ippacket = iphdr+segment
	return ippacket


