from crc import crc32DMR,crc_poly
from dmr_utils3.rs129 import encode
from dmr_utils3.decode import *
from dmr_utils3 import bptc,qr
from bitarray import bitarray
from crc import crc32DMR,crc_poly
import os

# Records and playback voice clips from DMR system
# Voice frames saved together with the MMDVM PDU, always 55 bytes

filenameStem = "VOICECLIP-"
baseTG = 1000 # Use three last digits of TG to record clip
dataLen = 55
f = 0
def createFile(id):
	global f
	filename = filenameStem + "%03d" % (id % baseTG)
	f = open(filename,"wb")

def appendVoiceData(data):
	global f
	if len(data)==dataLen:
		f.write(data)
	else:
		print ("Will not write data of length %d" % len(data))

def closeFile():
	global f
	f.close()

def getVoiceData(id,srcNR,dstNR):
	voiceClip = list()
	srcID = srcNR.to_bytes(3,"big")
	dstID = dstNR.to_bytes(3,"big")
	group = len(str(dstNR)) < 7 # ID less than 7 digits is a TG
	filename = filenameStem + "%03d" % (id % baseTG)
	if not os.path.isfile(filename):
		print ("Voice file does not exist")
		return list()
	f = open(filename,"rb")
	b = f.read(dataLen)
	while b:
		if b[15] & 0x30 == 0x20: # Data sync, modify LC
			dt = b[15] & 0xf # 1 - Voice header, 2, voice term
			b = modifyHeaderBlock(b,srcID,dstID,group,dt)
			#print("DataSync, ",b.hex())
		elif b[15] & 0x30 == 0: # Voice data
			b = modifyEmbeddedLC(b)
		# Replace DMR destination
		c = bytearray(b)
		c[5:8] = srcID
		c[8:11] = dstID
		if group: c[15] = c[15] & 0xbf
		else: c[15] = c[15] | 0x40
#		print ("-->",c.hex())
		voiceClip.append(c)
		b = f.read(dataLen)
	f.close()
	return voiceClip

# We have not been able to program the QR checksum, so we
# Insert a NULL embedded LC
def modifyEmbeddedLC(b):
	ba = bitarray()
	ba.frombytes(b)
	# Fill zerobits in positions 115-155
	ba[275:316] = 0
	return ba.tobytes()
#	parity1 = ba[108:116]
#	parity2 = ba[148:156]
#	print (parity1,parity2)
#	print (ba[116:148])
#	if seqnr == 1: embedLClist = bitarray() # New series of blocks
#	embedLClist.extend(ba[116:148])
#	if seqnr == 4: # Full LC received received
#		print ("Embedded LC length: ", len(embedLClist))
#		embed_lc = bptc.decode_emblc(embedLClist)
#		print (embed_lc)
#		embed_lc_coded = bptc.encode_emblc(embed_lc)
#		print (embed_lc_coded)
#		print (qr.encode(embed_lc))

# Replace header element with new addresses
def modifyHeaderBlock(b,srcID,dstID,group,datatype):
	bbit = bitarray()
	# Split the MMDVM PDU from DMR PDU
	mmdvmpdu = b[0:20]
	dmrpdu = b[20:53]
	bbit.frombytes(dmrpdu) # Convert to bitarray
	burst = data_burst(dmrpdu) # "Unpack"
	lc = burst['LC'] # Take out LC part, which we need to modify
	# Replace src and dest id, make new crc
#	print (lc.hex())
	if group:
		newLC = b'\0\0\0' + dstID + srcID
	else:
		newLC = b'\3\0\0' + dstID + srcID
	# Calculate new Reed Solomon CRC
	if datatype==1: mask = 0x969696
	elif datatype==2: mask = 0x999999
	else: print ("Illegal data type in modifyHeaderBlock")

	rs = encode(newLC)
	cs = (rs[0] << 16 | rs[1] << 8 | rs[2]) ^ mask
	# Append CRC value
	newLC += cs.to_bytes(3,"big")
#	print (newLC.hex())
	# Encode with turbo codes
	b0 = bptc.encode_19696(newLC)
	b1 = bptc.interleave_19696(b0)
	# Keep DT and SYNC parts,  append new LC
	bblock = b1[:98] + bbit[98:166] + b1[98:]
	return mmdvmpdu + bblock.tobytes() + b[-2:]

# Isolate the slot type and sync field in the burst, and decode the payload
# The burst format is described in fig.6.5 in ETSI ... document.
# The payload is 196 bits, split in two by a 20 bit SlotType field with
# a 48 bits sync field in the middle. The payload (called "LC" in the code)
# Is FEC-protected with Trellis coding. The payload must be unpacked, and the
# slot type fields extracted
def data_burst(_string):
#    	print(_string.hex())
	burst = to_bits(_string)
	info = burst[0:98] + burst[166:264]
	slot_type = burst[98:108] + burst[156:166]
	sync = burst[108:156] # Indicates the nature of the payload, see table 9.2 in ETSI ... document
	lc = xdecode_full_lc(info).tobytes() # Link Control, in our context this is IP fragments
	cc = to_bytes(slot_type[0:4])
	dtype = to_bytes(slot_type[4:8])
#	ret = {'LC': lc, 'CC': cc, 'DTYPE': dtype, 'SYNC': sync, 'STYPE':slot_type}
#	print(ret)
	return {'LC': lc, 'CC': cc, 'DTYPE': dtype, 'SYNC': sync, 'STYPE':slot_type}

# This is to decode the payload from the BPTC(96,196) FEC coding. Code taken from dmt_utils3 package
def xdecode_full_lc(_data):
    binlc = bitarray(endian='big')
    binlc.extend([_data[136],_data[121],_data[106],_data[91], _data[76], _data[61], _data[46], _data[31]])
    binlc.extend([_data[152],_data[137],_data[122],_data[107],_data[92], _data[77], _data[62], _data[47], _data[32], _data[17], _data[2]  ])
    binlc.extend([_data[123],_data[108],_data[93], _data[78], _data[63], _data[48], _data[33], _data[18], _data[3],  _data[184],_data[169]])
    binlc.extend([_data[94], _data[79], _data[64], _data[49], _data[34], _data[19], _data[4],  _data[185],_data[170],_data[155],_data[140]])
    binlc.extend([_data[65], _data[50], _data[35], _data[20], _data[5],  _data[186],_data[171],_data[156],_data[141],_data[126],_data[111]])
    binlc.extend([_data[36], _data[21], _data[6],  _data[187],_data[172],_data[157],_data[142],_data[127],_data[112],_data[97], _data[82] ])
    binlc.extend([_data[7],  _data[188],_data[173],_data[158],_data[143],_data[128],_data[113],_data[98], _data[83]])
    binlc.extend([ _data[68],_data[53],_data[174],_data[159],_data[144],_data[129],_data[114],_data[99],_data[84],_data[69],_data[54],_data[39],
    _data[24],_data[145],_data[130],_data[115],_data[100],_data[85],_data[70],_data[55],_data[40],_data[25],_data[10],_data[191]])
    return binlc



#l = getVoiceData(3,2420001,2420191)

#lc = bytes.fromhex("030000001f4124eddfa395e3")
#lc = bytes.fromhex("030000001f4324eddf")
#rs = encode(lc)
#rs = encode(lc[0:9])
#crc1 = crc_poly(lc[0:9],24,0x65b,xor_out=0x696969)
#print(rs)
#cs = (rs[0] << 16 | rs[1] << 8 | rs[2]) ^ 0x969696
#print(hex(cs))

#lc = bytes.fromhex("4d431ff114ee0850363097e0c4ad5d7f77fd757966382fd86930b8615d52c586ae002f")
#print(len(lc))
#b = data_burst(lc)
#print(b)
#print(b['LC'].hex())
